class ISCSIAuth(object):
    pass


class NoAuth(ISCSIAuth):
    def get_auth_name(self):
        return "NONE"


class ChapAuth(ISCSIAuth):
    def __init__(self, inbound_username, inbound_secret):
        self._inbound_username = inbound_username
        self._inbound_secret = inbound_secret

    def get_inbound_username(self):
        return self._inbound_username

    def get_inbound_secret(self):
        return self._inbound_secret

    def get_auth_name(self):
        return "CHAP"


class MutualChapAuth(ISCSIAuth):
    def __init__(self, inbound_username, inbound_secret, outbound_username, outbound_secret):
        self._inbound_username = inbound_username
        self._inbound_secret = inbound_secret
        self._outbound_username = outbound_username
        self._outbound_secret = outbound_secret

    def get_inbound_username(self):
        return self._inbound_username

    def get_inbound_secret(self):
        return self._inbound_secret

    def get_outbound_username(self):
        return self._outbound_username

    def get_outbound_secret(self):
        return self._outbound_secret

    def get_auth_name(self):
        return "MUTUAL_CHAP"



class Session(object):
    '''class that contains the iscsi session information
    '''
    def __init__(self, target, target_endpoint, source_ip, source_iqn, uid, hct):
        # TODO: add hct
        self._target = target
        self._target_endpoint = target_endpoint
        self._source_ip = source_ip
        self._source_iqn = source_iqn
        self._uid = uid
        self._hct = hct

    def __eq__(self, other):
        return isinstance(other, Session) and self._uid == other.get_uid() and self._target_endpoint == other.get_target_endpoint()

    def get_target(self):
        return self._target

    def get_uid(self):
        return self._uid

    def get_target_endpoint(self):
        return self._target_endpoint

    def get_source_ip(self):
        return self._source_ip

    def get_source_iqn(self):
        return self._source_iqn

    def get_hct(self):
        return self._hct


class Target(object):
    '''class that contains the iscsi connection spec
    '''
    def __init__(self, endpoints, discovery_endpoint, iqn):
        self._endpoints = endpoints
        self._discovery_endpoint = discovery_endpoint
        self._iqn = iqn

    def __eq__(self, other):
        return isinstance(other, Target) and self._iqn == other.get_iqn()

    def get_endpoints(self):
        return self._endpoints

    def get_discovery_endpoint(self):
        return self._discovery_endpoint

    def get_iqn(self):
        return self._iqn


class Initiator(object):
    '''class that contains the initiator iscsi details
    '''
    def __init__(self, iqn, initiator_name):
        self._iqn = iqn
        self._initiator_name = initiator_name

    def get_iqn(self):
        return self._iqn

    def get_initiator_name(self):
        return self._initiator_name


class Endpoint(object):
    '''Class that describes the iscsi target endpoint
    '''
    def __init__(self, ip_address, port):
        self._ip_address = ip_address
        self._port = port

    def __eq__(self, other):
        return isinstance(other, Endpoint) and self._ip_address == other.get_ip_address() and self._port == other.get_port()

    def get_ip_address(self):
        return self._ip_address

    def get_port(self):
        return self._port


class SoftwareInitiator(object):
    def is_installed(self):
        '''Platform specific function to understand if iSCSI can be used now
        '''
        raise NotImplementedError()

    def install(self):
        '''Platform specific function to active the iSCSI feature on an OS
        '''
        raise NotImplementedError()

    def uninstall(self):
        '''Platform specific function to deactivate iSCSI software
        '''
        raise NotImplementedError()


class ConnectionManager(object):
    '''Class that contains the main iscsi methods for connecting iscsi initiator
    to an iscsi target
    '''
    def discover(self, ip_address, port=3260):
        '''perform an iscsi discovery to an ip address
        '''
        raise NotImplementedError()

    def login(self, target, endpoint, auth=None, num_of_connections=1):
        '''receives target and endpoint and login to it
        '''
        raise NotImplementedError()

    def login_all(self, target, auth=None):
        '''login to all endpoints of a target and return the session it achieved
        '''
        raise NotImplementedError()

    def logout(self, session):
        '''receive a session and perform an iSCSI logout
        '''
        raise NotImplementedError()

    def logout_all(self, target):
        '''receive a target and logout of it
        '''
        raise NotImplementedError()

    def get_source_iqn(self):
        raise NotImplementedError()

    def reset_source_iqn(self):
        '''use incase iqn is invalid and regeneration of it is required'''
        raise NotImplementedError()

    def set_source_iqn(self, iqn):
        '''receive an iqn as a string, verify it's valid and set it.
        returns iqn type of the new IQN or None if fails
        '''
        raise NotImplementedError()

    def get_discovered_targets(self):
        '''return a list of discovered target objects
        '''
        raise NotImplementedError()

    def get_sessions(self, target=None):
        '''receive a target or None and return a list of all available sessions
        '''
        raise NotImplementedError()

    def rescan(self):
        '''rescan all available sessions
        '''
        raise NotImplementedError()

    def undiscover(self, target=None):
        '''delete all discovered sessions or only iqn specific active sessions
        '''
        raise NotImplementedError()



__iscsiapi = None
__iscsi_software_initiator = None


def get_iscsiapi():
    global __iscsiapi
    if __iscsiapi is None:
        __iscsiapi = _get_platform_specific_iscsiapi()
    return __iscsiapi


def get_iscsi_software_initiator():
    global __iscsi_software_initiator
    if __iscsi_software_initiator is None:
        __iscsi_software_initiator = _get_platform_specific_iscsi_software_initiator()
    return __iscsi_software_initiator


def _get_platform_specific_iscsiapi():
    from infi.os_info import get_platform_string
    platform = get_platform_string()
    if platform.startswith('windows'):
        from . import windows
        return windows.WindowsISCSIapi()
    elif platform.startswith('linux'):
        from . import linux
        return linux.LinuxISCSIapi()
    elif platform.startswith('solaris'):
        from . import solaris
        return solaris.SolarisISCSIapi()
    else:
        raise ImportError("not supported on this platform")


def _get_platform_specific_iscsi_software_initiator():
    from infi.os_info import get_platform_string
    platform = get_platform_string()
    if platform.startswith('windows'):
        from . import windows
        return windows.MicrosoftSoftwareInitiator()
    elif platform.startswith('linux'):
        from . import linux
        return linux.LinuxSoftwareInitiator()
    elif platform.startswith('solaris'):
        from . import solaris
        return solaris.SolarisSoftwareInitiator()
    else:
        raise ImportError("not supported on this platform")



from . import base
from . import auth as iscsiapi_auth
from infi.dtypes.iqn import IQN
from infi.os_info import get_platform_string
import infi.pkgmgr
import os

from logging import getLogger
logger = getLogger(__name__)

if 'ubuntu' in get_platform_string() or 'suse' in get_platform_string():
    ISCSI_CONNECTION_CONFIG = '/etc/iscsi/nodes'
else:
    ISCSI_CONNECTION_CONFIG = '/var/lib/iscsi/nodes'
ISCSI_INITIATOR_IQN_FILE = '/etc/iscsi/initiatorname.iscsi'
GENERATE_COMMAND = 'iscsi-iname'

class LinuxISCSIapi(base.ConnectionManager):

    def _execute(self, cmd):
        from infi.execute import execute
        logger.debug("Running: {}".format(cmd))
        return execute(cmd)

    def _execute_assert_success(self, cmd):
        from infi.execute import execute_assert_success
        logger.debug("Running: {}".format(cmd))
        return execute_assert_success(cmd)

    def _parse_discovery_endpoint(self, iqn):
        '''get an iqn of discovered target and return the discovery ip address
        '''
        import re
        IQN(iqn)  # make sure it's valid iqn
        for end_point in os.listdir(os.path.join(ISCSI_CONNECTION_CONFIG, iqn)):
            filepath = os.path.join(ISCSI_CONNECTION_CONFIG, iqn, end_point)
            # HPT-2193 filepath could be a file with the node info, or a dir that contains a file "default" which
            # has the information
            if os.path.isdir(filepath):
                filepath = os.path.join(filepath, 'default')
            if not os.path.isfile(filepath):
                continue
            try:
                with open(filepath, 'r') as fd:
                    content = fd.read()
            except (OSError, IOError):
                continue
            ip = re.search("node.discovery_address\s*=\s*(\d+\.\d+\.\d+\.\d+)", content)
            port = re.search("node.discovery_port\s*=\s*(\d+)", content)
            if ip is not None and port is not None:
                return base.Endpoint(ip.group(1), int(port.group(1)))

    def get_discovered_targets(self):
        targets = []
        if not os.path.isdir(ISCSI_CONNECTION_CONFIG):
            return targets
        for iqn in os.listdir(ISCSI_CONNECTION_CONFIG):
            endpoints = []
            for end_point in os.listdir(os.path.join(ISCSI_CONNECTION_CONFIG, iqn)):
                dst_ip, dst_port = end_point.split(",")[:2]
                endpoints.append(base.Endpoint(dst_ip, int(dst_port)))
            discovery_endpoint = self._parse_discovery_endpoint(iqn)
            # HPT-2193 discovery_endpoint could be None, we must not fail because of this
            targets.append(base.Target(endpoints, discovery_endpoint, iqn))
        return targets

    def _iter_sessions_in_sysfs(self):
        import re
        from infi.dtypes.hctl import HCT
        from glob import glob

        def sysfs_file_content(path):
            with open(path, 'r') as fd:
                return fd.read().rstrip(' \t\r\n\0')

        for host in glob(os.path.join('/sys', 'devices', 'platform', 'host*')):

            # some older versions of RHEL-based operating systems use the former path variant, others use the latter
            iscsi_host = glob(os.path.join(host, 'iscsi_host*host*')) + \
                         glob(os.path.join(host, 'iscsi_host', 'host*'))
            source_ip = sysfs_file_content(os.path.join(iscsi_host[0], 'ipaddress'))

            for session in glob(os.path.join(host, 'session*')):  # usually, one session per host
                uid = re.split('^session', os.path.basename(session))[1]

                iscsi_session = glob(os.path.join(session, 'iscsi_session*session*')) + \
                                glob(os.path.join(session, 'iscsi_session', 'session*'))
                source_iqn = sysfs_file_content(os.path.join(iscsi_session[0], 'initiatorname'))
                target_iqn = sysfs_file_content(os.path.join(iscsi_session[0], 'targetname'))

                connections = glob(os.path.join(session, 'connection*', 'iscsi_connection*connection*')) + \
                              glob(os.path.join(session, 'connection*', 'iscsi_connection', 'connection*'))

                for connection in connections:
                    try:
                        ip_address = sysfs_file_content(os.path.join(connection, 'address'))
                        port = sysfs_file_content(os.path.join(connection, 'port'))
                        break
                    except (IOError, OSError):
                        logger.debug("connection parameters are missing for {}".format(connection))
                        continue
                else:  # no connections in session
                    logger.debug("no valid connection for session {}".format(session))
                    continue

                target_endpoint = base.Endpoint(ip_address, int(port))

                for target in glob(os.path.join(session, 'target*')):  # usually, one target per session
                    target_id = os.path.basename(target)
                    hct_tuple = re.split('^target', target_id)[1].split(':')
                    hct = HCT(*(int(i) for i in hct_tuple))
                    break
                else:  # no targets in session
                    logger.debug("no targets for session {}".format(session))
                    continue

                yield target_iqn, target_endpoint, source_ip, source_iqn, uid, hct

    def _get_sessions_from_sysfs(self):
        sessions = []
        iqn_to_target = {target.get_iqn(): target for target in self.get_discovered_targets()}

        for target_iqn, target_endpoint, source_ip, source_iqn, uid, hct in self._iter_sessions_in_sysfs():
            if target_iqn not in iqn_to_target:
                # no matching targets
                # TODO should we raise an exception here? Or pass target=None to base.Session?
                # We saw in HPT-2193 that skipping sessions may be a big problem because it means we may miss iSCSI
                # devices and later treat them as local devices instead
                msg = "no valid target for session in sysfs. target endpoint={}:{} target iqn={}"
                logger.debug(msg.format(target_endpoint.get_ip_address(), target_endpoint.get_port(), target_iqn))
                continue
            target = iqn_to_target[target_iqn]
            session = base.Session(target, target_endpoint, source_ip, source_iqn, uid, hct)
            sessions.append(session)
        return sessions

    def _reload_iscsid_service(self):
        if os.path.isfile('/bin/systemctl'):
            self._execute_assert_success(['systemctl', 'restart', 'iscsid'])
        elif os.path.isfile('/etc/init.d/iscsid'):
            self._execute_assert_success(['service', 'iscsid', 'restart'])
        elif os.path.isfile('/etc/init.d/open-iscsi'):
            self._execute_assert_success(['service', 'open-iscsi', 'restart'])
        else:
            logger.error("couldn't find /bin/systemctl or /usr/sbin/service, failed reloading")

    def _remove_comments(self, list_of_strings):
        '''get list of strings and return list of strings without the commented out ones'''
        import re
        no_comment = []
        regex = re.compile(r'^#.+')
        for line in list_of_strings:
            if not re.match(regex, line):
                no_comment.append(line)
        return no_comment

    def _update_node_parameter(self, name, value, target):
        from infi.execute import execute_assert_success
        args = ['iscsiadm', '-m', 'node', '-o', 'update', '-n', name, '-v', value, '-T', target]
        if "password" not in name:
            logger.debug("running {}".format(args))
        else:
            logger.debug("running {}".format(args).replace(value, '***' if value else ''))
        return execute_assert_success(args)  # we don't want to log the password

    def _set_auth(self, auth, target):
        target_iqn = target.get_iqn()
        if isinstance(auth, iscsiapi_auth.ChapAuth):
            self._update_node_parameter('node.session.auth.authmethod', 'CHAP', target_iqn)
            self._update_node_parameter('node.session.auth.username', auth.get_inbound_username(), target_iqn)
            self._update_node_parameter('node.session.auth.password', auth.get_inbound_secret(), target_iqn)
            self._update_node_parameter('node.session.auth.username_in', '', target_iqn)
            self._update_node_parameter('node.session.auth.password_in', '', target_iqn)
        elif isinstance(auth, iscsiapi_auth.MutualChapAuth):
            self._update_node_parameter('node.session.auth.authmethod', 'CHAP', target_iqn)
            self._update_node_parameter('node.session.auth.username', auth.get_inbound_username(), target_iqn)
            self._update_node_parameter('node.session.auth.password', auth.get_inbound_secret(), target_iqn)
            self._update_node_parameter('node.session.auth.username_in', auth.get_outbound_username(), target_iqn)
            self._update_node_parameter('node.session.auth.password_in', auth.get_outbound_secret(), target_iqn)
        elif isinstance(auth, iscsiapi_auth.NoAuth):
            self._update_node_parameter('node.session.auth.authmethod', 'None', target_iqn)
            self._update_node_parameter('node.session.auth.username', "", target_iqn)
            self._update_node_parameter('node.session.auth.password', "", target_iqn)
            self._update_node_parameter('node.session.auth.username_in', "", target_iqn)
            self._update_node_parameter('node.session.auth.password_in', "", target_iqn)

    def _get_old_iqn(self):
        from os.path import isfile
        if not isfile(ISCSI_INITIATOR_IQN_FILE):
            return
        with open(ISCSI_INITIATOR_IQN_FILE, 'r') as fd:
            return fd.readlines()

    def _session_already_active(self, target, target_endpoint):
        for session in self.get_sessions(target=target):
            if target_endpoint == session.get_target_endpoint():
                return session

    def get_source_iqn(self):
        '''return infi.dtypes.iqn type iqn if iscsi initiator file exists
        '''
        from .iscsi_exceptions import NotReadyException
        import re
        from os.path import isfile
        if not isfile(ISCSI_INITIATOR_IQN_FILE):
            raise NotReadyException("iSCSI initiator IQN file not found")
        with open(ISCSI_INITIATOR_IQN_FILE, 'r') as fd:
            data = self._remove_comments(fd.readlines())
            assert len(data) == 1, "something isn't right with {}".format(ISCSI_INITIATOR_IQN_FILE)
            raw_iqn = re.split('InitiatorName=', data[0])
            return IQN(raw_iqn[1].strip())

    def reset_source_iqn(self):
        '''use in case iqn is invalid and regeneration of it is required'''
        process = self._execute_assert_success([GENERATE_COMMAND])
        iqn = process.get_stdout().strip()
        IQN(iqn)  # validating new IQN
        logger.info("Regeneration of iqn was initiated, old file {}".format(ISCSI_INITIATOR_IQN_FILE) +
                    "had this data in it {!r}, new iqn is:{}".format(self._get_old_iqn(), iqn))
        self.set_source_iqn(iqn)

    def set_source_iqn(self, iqn):
        '''receives a string, validates it's an iqn then set it to the host
        NOTE: this restart the iscsi service and may fail active sessions !
        '''
        IQN(iqn)   # checks iqn is valid
        logger.info("Old IQN was:{}".format(self._get_old_iqn()))
        replacement_strig = 'InitiatorName=' + iqn
        with open(ISCSI_INITIATOR_IQN_FILE, 'w') as fd:
            fd.write(replacement_strig + "\n")
        logger.info("iqn was replaced to {}".format(iqn))
        self._reload_iscsid_service()

    def discover(self, ip_address, port=3260):
        '''initiate discovery and returns a list of dicts which contain all available targets
        '''
        args = ['iscsiadm', '-m', 'discovery', '-t', 'st', '-p', str(ip_address) + ':' + str(port)]
        logger.info("running {}".format(args))
        self._execute_assert_success(args)

        discovery_endpoint = base.Endpoint(ip_address, port)
        targets = self.get_discovered_targets()
        return [target for target in targets if discovery_endpoint in target.get_endpoints()][0]

    def login(self, target, endpoint, auth=None, num_of_connections=1):
        if auth is None:
            auth = iscsiapi_auth.NoAuth()
        self._set_auth(auth, target)
        if self._session_already_active(target, endpoint) is None:
            args = ['iscsiadm', '-m', 'node', '-l', '-T', target.get_iqn(), '-p',
            "{}:{}".format(endpoint.get_ip_address(), endpoint.get_port())]
            self._execute_assert_success(args)
        for session in self._get_sessions_from_sysfs():
            if session.get_target_endpoint() == endpoint:
                return session

    def login_all(self, target, auth=None):
        for endpoint in target.get_endpoints():
            self.login(target, endpoint, auth)
        return self.get_sessions(target=target)

    def logout(self, session):
        ip_address = session.get_target_endpoint().get_ip_address()
        self._execute((['iscsiadm', '-m', 'node', '-u', '-T', str(session.get_target().get_iqn()), '-p', ip_address]))

    def logout_all(self, target):
        self._execute((['iscsiadm', '-m', 'node', '-u', '-T', str(target.get_iqn())]))

    def get_sessions(self, target=None):
        '''receive a target or None and return a list of all available sessions
        '''
        sessions = self._get_sessions_from_sysfs()
        if not target:
            return sessions
        return [session for session in sessions if session.get_target_endpoint() in target.get_endpoints()]

    def rescan(self):
        '''rescan all available sessions
        '''
        self._execute(['iscsiadm', '-m', 'session', '--rescan'])

    def undiscover(self, target=None):
        '''logout from everything and delete all discovered target if target=None otherwise delete only the target
        discovery endpoints
        '''
        if target:
            self.logout_all(target)
            self._execute(['iscsiadm', '-m', 'node', '-o', 'delete', str(target.get_iqn())])
        else:
            for target in self.get_discovered_targets():
                self.logout_all(target)
            self._execute(['iscsiadm', '-m', 'node', '-o', 'delete'])


class LinuxSoftwareInitiator(base.SoftwareInitiator):
    def is_installed(self):
        ''' In linux, return True if iSCSI initiator sw is installed otherwise return False
        '''
        if 'centos' in get_platform_string() or 'redhat' in get_platform_string():
            pkgmgr = infi.pkgmgr.get_package_manager()
            return pkgmgr.is_package_installed('iscsi-initiator-utils')
        if 'ubuntu' in get_platform_string() or 'suse' in get_platform_string():
            pkgmgr = infi.pkgmgr.get_package_manager()
            return pkgmgr.is_package_installed('open-iscsi')

    def install(self):
        if 'centos' in get_platform_string() or 'redhat' in get_platform_string():
            pkgmgr = infi.pkgmgr.get_package_manager()
            pkgmgr.install_package('iscsi-initiator-utils')
        if 'ubuntu' in get_platform_string() or 'suse' in get_platform_string():
            pkgmgr = infi.pkgmgr.get_package_manager()
            pkgmgr.install_package('open-iscsi')
            if 'suse-12' in get_platform_string():
                self._execute(['service', 'iscsid', 'start'])

    def uninstall(self):
        if 'centos' in get_platform_string() or 'redhat' in get_platform_string():
            pkgmgr = infi.pkgmgr.get_package_manager()
            pkgmgr.remove_package('iscsi-initiator-utils')
        if 'ubuntu' in get_platform_string() or 'suse' in get_platform_string():
            pkgmgr = infi.pkgmgr.get_package_manager()
            pkgmgr.remove_package('open-iscsi')



from infi.win32service import ServiceControlManagerContext
from infi.execute import execute_assert_success, execute, ExecutionError
from . import base
from . import auth as iscsiapi_auth
from infi.dtypes.iqn import IQN
from infi.wmi import WmiClient

from logging import getLogger
logger = getLogger(__name__)

ISCSI_LOGIN_FLAG_MULTIPATH_DISABLED = 0
ISCSI_LOGIN_FLAG_REQUIRE_IPSEC = 1
ISCSI_LOGIN_FLAG_MULTIPATH_ENABLED = 2
ISCSI_NO_AUTH_TYPE = 0
ISCSI_CHAP_AUTH_TYPE = 1
ISCSI_MUTUAL_CHAP_AUTH_TYPE = 2
ISCSI_SECURITY_TUNNEL_MODE_PREFERRED = 0x00000040
ISCSI_SECURITY_TRANSPORT_MODE_PREFERRED = 0X00000020
ISCSI_SECURITY_PFS = 0x00000010
ISCSI_SECURITY_NEGOTIATE_VIA_AGGRESSIVE_MODE = 0x00000008
ISCSI_SECURITY_NEGOTIATE_VIA_MAIN_MODE = 0x00000004
ISCSI_SECURITY_IPSEC_ENABLED = 0x00000002
ISCSI_SECURITY_VALID_FLAGS = 1


class WindowsISCSIapi(base.ConnectionManager):
    def __init__(self):
        super(WindowsISCSIapi, self).__init__()
        self._initiator = None
        self._login_flags = ISCSI_LOGIN_FLAG_MULTIPATH_ENABLED
        self._security_flags = 0

    def disable_mpio(self):
        self._login_flags = ISCSI_LOGIN_FLAG_MULTIPATH_DISABLED

    def enable_mpio(self):
        self._login_flags = ISCSI_LOGIN_FLAG_MULTIPATH_ENABLED

    def _create_initiator_obj_if_needed(self):
        if not self._initiator:
            iscsi_software = MicrosoftSoftwareInitiator()
            if not iscsi_software.is_installed():
                raise RuntimeError("iscsi sw isn't running")
            client = WmiClient('root\\wmi')
            query = client.execute_query('SELECT * from MSIscsiInitiator_InitiatorClass')
            initiator_name = list(query)[0].Properties_.Item('InitiatorName').Value
            iqn = self.get_source_iqn()
            self._initiator = base.Initiator(iqn, initiator_name)

    def _refresh_wmi_db(self):
        client = WmiClient('root\\wmi')
        for item in client.execute_query("SELECT * FROM MSIscsiInitiator_MethodClass"):
            item.ExecMethod_('RefreshTargetList', None)

    def _execute_discover(self, ip_address, port):
        from .iscsi_exceptions import DiscoveryFailed
        try:
            execute_assert_success(['iscsicli', 'AddTargetPortal', str(ip_address), str(port)])
        except ExecutionError as e:
            msg = "couldn't connect to ip_address {!r}, error: {!r}" + \
                  "This could be due to one of the following reasons:" + \
                  "1. There is no IP connectivity between your host and {!r}" + \
                  "2. The iSCSI server is down"
            formated_msg = msg.format(ip_address, e, ip_address)
            logger.error(formated_msg)
            raise DiscoveryFailed(formated_msg)

    def _return_target(self, ip_address, port):
        endpoints = []
        iqn = None
        sessions = self._get_connectivity_using_wmi()
        for session in sessions:
            if session['dst_ip'] == ip_address:
                iqn = IQN(session['iqn'])
        if iqn is None:
            raise RuntimeError("iqn is empty, it means that the discovery address {} didn't returned from the target"
                               .format(ip_address))
        for session in sessions:
            if session['iqn'] == iqn:
                endpoints.append(base.Endpoint(session['dst_ip'], session['dst_port']))
        return base.Target(endpoints, base.Endpoint(ip_address, port), iqn)

    def _get_discovery_endpoints(self):
        '''return all discovery endpoints currently use only for undiscover
        '''
        discovery_endpoints = []
        client = WmiClient('root\\wmi')
        query = client.execute_query('SELECT * FROM  MSiSCSIInitiator_SendTargetPortalClass')
        for discovery_endpoint in query:
            endpoint = base.Endpoint(discovery_endpoint.Properties_.Item('PortalAddress').Value,
            discovery_endpoint.Properties_.Item('PortalPort').Value)
            if endpoint not in discovery_endpoints:
                discovery_endpoints.append(endpoint)
        return discovery_endpoints

    def _return_auth_type(self, auth):
        if isinstance(auth, iscsiapi_auth.ChapAuth):
            return ISCSI_CHAP_AUTH_TYPE
        if isinstance(auth, iscsiapi_auth.MutualChapAuth):
            return ISCSI_MUTUAL_CHAP_AUTH_TYPE
        if isinstance(auth, iscsiapi_auth.NoAuth):
            return ISCSI_NO_AUTH_TYPE

    def _remove_persistent_target(self, target):
        self._create_initiator_obj_if_needed()
        for endpoint in target.get_endpoints():
            args = ['iscsicli', 'RemovePersistentTarget', str(self._initiator.get_initiator_name()), str(target.get_iqn()),
                                    '*', str(endpoint.get_ip_address()), str(endpoint.get_port())]
            logger.info("running {}".format(args))
            execute(args)

    def _reomve_target_portal(self, target):
        for endpoint in target.get_endpoints():
            args = ['iscsicli', 'RemoveTargetPortal', str(endpoint.get_ip_address()), str(endpoint.get_port())]
            logger.info("running {}".format(args))
            execute(args)

    def discover(self, ip_address, port=3260):
        '''perform an iscsi discovery to an ip address
        '''
        # TODO: support chap
        self._create_initiator_obj_if_needed()
        already_discoverd = False
        discovery_endpoint = base.Endpoint(ip_address, port)
        for target in self.get_discovered_targets():
            if target.get_discovery_endpoint() == discovery_endpoint:
                self._refresh_wmi_db()
                already_discoverd = True
                break
        if not already_discoverd:
            self._execute_discover(ip_address, port)
        return self._return_target(ip_address, port)

    def login(self, target, endpoint, auth=None, num_of_connections=1):
        '''receives target and endpoint and login to it
        '''
        # LoginTarget is not persistent across reboots
        # PersistentLoginTarget will make sure we connect after reboot but not immediately
        # so we need to call both
        if auth is None:
            auth = iscsiapi_auth.NoAuth()
        self._iscsicli_login('LoginTarget', target, endpoint, auth, num_of_connections)
        self._iscsicli_login('PersistentLoginTarget', target, endpoint, auth, num_of_connections)
        for session in self.get_sessions():
            if session.get_target_endpoint() == endpoint:
                return session

    def _iscsicli_login(self, login_command, target, endpoint, auth=None, num_of_connections=1):

        auth_type = self._return_auth_type(auth)
        if auth_type == ISCSI_CHAP_AUTH_TYPE:
            username = auth.get_inbound_username()
            password = auth.get_inbound_secret()
        elif auth_type == ISCSI_MUTUAL_CHAP_AUTH_TYPE:
            username = auth.get_inbound_username()
            password = auth.get_inbound_secret()
            cmd = ['iscsicli', 'CHAPSecret', auth.get_outbound_secret()]
            execute(cmd)
        elif auth_type == ISCSI_NO_AUTH_TYPE:
            username = '*'
            password = '*'

        # Due to a bug only in 2008 multiple sessions isn't handled ok unless initiator name is monitored
        # Therefore we don't use Qlogin, Details:
        # https://social.technet.microsoft.com/Forums/office/en-US/4b2420d6-0f28-4d12-928d-3920896f582d/iscsi-initiator-target-not-reconnecting-on-reboot?forum=winserverfiles
        # http://download.microsoft.com/download/a/e/9/ae91dea1-66d9-417c-ade4-92d824b871af/uguide.doc

        # iscsicli LoginTarget <TargetName> <ReportToPNP>
        #                      <TargetPortalAddress> <TargetPortalSocket>
        #                      <InitiatorInstance> <Port number> <Security Flags>
        #                     <Login Flags> <Header Digest> <Data Digest>
        #                     <Max Connections> <DefaultTime2Wait>
        #                     <DefaultTime2Retain> <Username> <Password> <AuthType> <Key>
        #                     <Mapping Count> <Target Lun> <OS Bus> <Os Target>
        #                     <OS Lun> ...

        # iscsicli PersistentLoginTarget <TargetName> <ReportToPNP>
        #                      <TargetPortalAddress> <TargetPortalSocket>
        #                     <InitiatorInstance> <Port number> <Security Flags>
        #                     <Login Flags> <Header Digest> <Data Digest>
        #                     <Max Connections> <DefaultTime2Wait>
        #                     <DefaultTime2Retain> <Username> <Password> <AuthType> <Key>
        #                     <Mapping Count> <Target Lun> <OS Bus> <Os Target>
        #                     <OS Lun> ...
        command = '''iscsicli {0} {TargetName} {ReportToPNP}
                {TargetPortalAddress} {TargetPortalSocket}
                {InitiatorInstance} {Port_number} {Security_Flags}
                {Login_Flags} {Header_Digest} {Data_Digest}
                {Max_Connections} {DefaultTime2Wait}
                {DefaultTime2Retain} {Username} {Password} {AuthType} {Key}
                {Mapping_Count}'''

        args = command.format(login_command, TargetName=target.get_iqn(),
                              ReportToPNP='t',  # If the value is T or t then the LUN is exposed as a device
                              TargetPortalAddress=endpoint.get_ip_address(),
                              TargetPortalSocket=endpoint.get_port(),
                              InitiatorInstance=self._initiator.get_initiator_name(),
                              Port_number='*',  # the kernel mode initiator driver chooses the initiator port used
                              Security_Flags=self._security_flags,
                              Login_Flags=self._login_flags,
                              Header_Digest='*',  # the digest setting is determined by the initiator kernel mode driver
                              Data_Digest=0,
                              Max_Connections='*',  # the kernel mode initiator driver chooses the value for maximum connections
                              DefaultTime2Wait=0,
                              DefaultTime2Retain=0,
                              Username=username,  # the iSCSI initiator service will use the initiator node name as the CHAP username
                              Password=password,  # The initiator will use this secret to compute a hash value based on the challenge sent by the target
                              AuthType=auth_type,
                              Key=0,
                              Mapping_Count=0)
        logger.info("running iscsicli LoginTarget {!r}".format(args))
        process = execute(args.split())
        if int(process.get_returncode()) != 0:
            logger.info("couldn't login to {!r} {!r} {!r} because: {!r}"
                        .format(target.get_iqn(), endpoint.get_ip_address(), endpoint.get_port(), process.get_stdout()))
            if "target has already been logged in" not in process.get_stdout():
                raise RuntimeError(process.get_stdout())

    def login_all(self, target, auth=None):
        ''' login to all endpoint of a target and return the session it achieved
        '''
        if auth is None:
            auth = iscsiapi_auth.NoAuth()
        for endpoint in target.get_endpoints():
            self.login(target, endpoint, auth)
        return self.get_sessions(target=target)

    def logout(self, session):
        '''receive a session and perform an iSCSI logout
        '''
        execute_assert_success(['iscsicli', 'LogoutTarget', str(session.get_uid())])

    def logout_all(self, target):
        '''receive a target and logout of it
        '''
        for session in self.get_sessions(target):
            self.logout(session)

    def get_source_iqn(self):
        from .iscsi_exceptions import NotReadyException
        client = WmiClient('root\\wmi')
        query = list(client.execute_query('SELECT * FROM MSIscsiInitiator_MethodClass'))
        if not query:
            raise NotReadyException("Could not query iSCSI initiator from WMI")
        iqn = query[0].Properties_.Item("ISCSINodeName").Value
        return IQN(iqn)

    def reset_source_iqn(self):
        execute_assert_success(['iscsicli', 'NodeName', '*'])

    def set_source_iqn(self, iqn):
        '''receive an iqn as a string, verify it's valid and set it.
        returns iqn type of the new IQN or None if fails
        '''
        logger.info("iqn before the change is {!r} going to change to {!r}".format(self.get_source_iqn(), iqn))
        IQN(iqn)  # raise if iqn doesn't exist
        client = WmiClient('root\\wmi')
        query = list(client.execute_query("SELECT * FROM MSIscsiInitiator_MethodClass"))[0]
        initiator_name = query.Methods_.Item("SetIscsiInitiatorNodeName")
        parameters = initiator_name.InParameters.SpawnInstance_()
        parameters.Properties_.Item("InitiatorNodeName").Value = iqn
        query.ExecMethod_('SetIscsiInitiatorNodeName', parameters)
        logger.info("iqn is now {!r}".format(self.get_source_iqn()))
        return self.get_source_iqn()

    def _get_connectivity_using_wmi(self):
        '''returns a list of dicts which contain all available targets with it's main parameters
        '''
        availble_targets_connectivity = []
        client = WmiClient('root\\wmi')
        for target in client.execute_query('SELECT * from  MSIscsiInitiator_TargetClass'):
            iqn = target.Properties_.Item('TargetName').Value
            for portal_group in target.Properties_.Item('PortalGroups').Value:
                for portal in portal_group.Properties_.Item('Portals').Value:
                    target_connectivity = {'dst_ip': portal.Properties_.Item('Address').Value,
                                           'dst_port': portal.Properties_.Item('Port').Value, 'iqn': iqn}
                    if target_connectivity not in availble_targets_connectivity:
                        availble_targets_connectivity.append(target_connectivity)
        return availble_targets_connectivity

    def get_discovered_targets(self):
        '''return a list of discovered target objects
        '''
        logger.info("get_discovered_targets")
        discovered_targets = []
        client = WmiClient('root\\wmi')
        for query in client.execute_query('SELECT * from MSIscsiInitiator_TargetClass'):
            endpoints = []
            iqn = query.Properties_.Item('TargetName').Value
            for portal_group in query.Properties_.Item('PortalGroups').Value:
                for portal in portal_group.Properties_.Item('Portals').Value:
                    endpoint = base.Endpoint(portal.Properties_.Item('Address').Value, portal.Properties_.Item('Port').Value)
                    if endpoint not in endpoints:
                        endpoints.append(endpoint)

            # u'SendTargets:*test 0003260 ROOT\\ISCSIPRT\\0000_0 '
            # HOSTDEV-2764: u'SendTargets:*172.16.130.9 0003260 ROOT\\ISCSIPRT\\0000_0 172.16.130.55'
            discovery_mechanism = query.Properties_.Item('DiscoveryMechanism').Value
            if not discovery_mechanism:
                logger.debug("invalid discovery mechanism (empty): {!r}".format(discovery_mechanism))
                continue
            discovery_mechanism = discovery_mechanism.strip().split()
            if len(discovery_mechanism) < 2:
                logger.debug("invalid discovery mechanism: {!r}".format(discovery_mechanism))
                continue
            discovery_endpoint_ip = discovery_mechanism[0].split(':')[-1].strip("*")
            discovery_endpoint_port = int(discovery_mechanism[1])
            discovery_endpoint = base.Endpoint(discovery_endpoint_ip, discovery_endpoint_port)
            target = base.Target(endpoints, discovery_endpoint, iqn)
            if target not in discovered_targets:
                discovered_targets.append(target)
        return discovered_targets

    def get_sessions(self, target=None):
        '''receive a target or None and return a list of all available sessions
        '''
        # assumes only one connection over each session ( conn_0 ) 3.0 Infinibox limit
        # TODO: when InfiniBox will support MCS need to modify this code
        from infi.dtypes.hctl import HCT
        logger.info("get_sessions(target={!r}".format(target))

        def _get_sessions_of_target(target):
            client = WmiClient('root\\wmi')
            wql = "SELECT * from MSiSCSIInitiator_SessionClass where TargetName='%s'" % str(target.get_iqn())
            query = client.execute_query(wql)
            target_sessions = []
            for session in query:
                hct = None
                uid = session.Properties_.Item('SessionId').Value
                if not uid:
                    continue
                connections = session.Properties_.Item('ConnectionInformation').Value
                if not connections:
                    continue

                conn_0 = connections[0]
                source_ip = conn_0.Properties_.Item('InitiatorAddress').Value
                source_iqn = session.Properties_.Item('InitiatorName').Value
                target_address = conn_0.Properties_.Item('TargetAddress').Value
                target_port = conn_0.Properties_.Item('TargetPort').Value

                if not session.Properties_.Item('Devices').Value:
                    hct = HCT(-1, 0, -1)
                else:
                    devices = list(session.Properties_.Item('Devices').Value)
                    hct = HCT(devices[0].Properties_.Item('ScsiPortNumber').Value,
                              devices[0].Properties_.Item('ScsiPathId').Value,
                              devices[0].Properties_.Item('ScsiTargetId').Value)
                target_sessions.append(base.Session(target, base.Endpoint(target_address, target_port), source_ip, source_iqn, uid, hct))
            return target_sessions

        if target:
            return _get_sessions_of_target(target)
        else:
            sessions = []
            for target in self.get_discovered_targets():
                sessions.extend(_get_sessions_of_target(target))
            return sessions

    def rescan(self):
        '''rescan all available sessions
        '''
        self._refresh_wmi_db()

    def undiscover(self, target=None):
        '''logout and delete all discovered sessions for a target or for all targets
        '''
        if target:
            self.logout_all(target)
            self._reomve_target_portal(target)
            self._remove_persistent_target(target)
        else:
            for target in self.get_discovered_targets():
                self.logout_all(target)
                self._reomve_target_portal(target)
                self._remove_persistent_target(target)
        self._refresh_wmi_db()

class MicrosoftSoftwareInitiator(base.SoftwareInitiator):
    def is_installed(self):
        '''in windows return True if iSCSI initiator sw is running otherwise return False
        '''
        with ServiceControlManagerContext() as scm:
            with scm.open_service('MSiSCSI') as service:
                return service.is_running() and service.is_autostart()

    def install(self):
        '''start the iSCSI service on windows.
           in the future will also autostart the service
        '''
        logger.debug("trying to start service MSiSCSI")
        with ServiceControlManagerContext() as scm:
            with scm.open_service('MSiSCSI') as service:
                logger.debug("starting service MSiSCSI")
                service.start_automatically()
                service.safe_start()
                service.wait_on_pending()

    def uninstall(self):
        '''Stop the iscsi service on windows
        '''
        with ServiceControlManagerContext() as scm:
            with scm.open_service('MSiSCSI') as service:
                logger.debug("stopping MSiSCSI")
                service.safe_stop()
                service.wait_on_pending()
                service.disable()



__version__ = "0.2.9"
__git_commiter_name__ = "Arnon Yaari"
__git_commiter_email__ = "arnony@infinidat.com"
__git_branch__ = u'(HEAD detached at v0.2.9)'
__git_remote_tracking_branch__ = '(No remote tracking)'
__git_remote_url__ = '(Not remote tracking)'
__git_head_hash__ = '75cb6806d8ae66c1c07be363dd6636b9c6607c61'
__git_head_subject__ = u'Finished Release v0.2.9'
__git_head_message__ = u''
__git_dirty_diff__ = u''
__git_commit_date__ = '2018-02-06 16:25:26'



from infi.execute import execute_assert_success, execute
from . import auth as iscsiapi_auth
from . import base
from infi.dtypes.iqn import IQN
from infi.os_info import get_platform_string

from logging import getLogger
logger = getLogger(__name__)

class SolarisISCSIapi(base.ConnectionManager):
    def _execute_assert_n_log(self, cmd, log_prefix='running: ', log_level='debug'):
        try:
            getattr(logger, str(log_level))(log_prefix + "{}".format(cmd if isinstance(cmd, basestring) else ' '.join(cmd)))
        except AttributeError as e:
            logger.error("logger.{} doesn't exist, {!r}".format(log_level, e))
        return execute_assert_success(cmd)

    def _execute_n_log(self, cmd, log_prefix='running: ', log_level='debug'):
        try:
            getattr(logger, str(log_level))(log_prefix + (cmd if isinstance(cmd, basestring) else ' '.join(cmd)))
        except AttributeError as e:
            logger.error("logger.{} doesn't exist, {!r}".format(log_level, e))
        return execute(cmd)

    def _set_number_of_connection_to_infinibox(self):
        '''In Solaris we need to configure in advance how many session an initiator can open.
        This to each target
        '''
        #  Not being used anymore due to INFINIBOX-24755
        con_number = self._how_many_connections_should_be_configured()
        logger.info("Changing number of sessions to {}".format(con_number))
        cmd = ['iscsiadm', 'modify', 'initiator-node', '-c', str(con_number)]
        self._execute_assert_n_log(cmd)

    def _how_many_connections_should_be_configured(self):
        # needs rewrite if used ( 4.0 ), need to verify that first target is infinibox()
        max_endpoints = len(self.get_discovered_targets()[0].get_endpoints())
        for target in self.get_discovered_targets():
            if max_endpoints < len(target.get_endpoints()):
                max_endpoints = len(target.get_endpoints())
        return max_endpoints

    def _parse_discovered_targets(self):
        import re
        availble_targets = []
        cmd = ['iscsiadm', 'list', 'discovery-address']
        process = self._execute_assert_n_log(cmd)
        if len(list(process.get_stdout())) == 0:
            return availble_targets
        cmd = ['iscsiadm', 'list', 'discovery-address', '-v']
        process = self._execute_n_log(cmd)
        if process.get_returncode() != 0:
            return availble_targets
        output = process.get_stdout().splitlines()
        for line_number, line in enumerate(output):
            if re.search(r'Target name:', line):
                if re.search(r'Target address:', output[line_number + 1]):
                    regex = re.compile(r'(?P<dst_ip>\d+\.\d+\.\d+\.\d+)\:(?P<dst_port>\d+)')
                    session = regex.search(output[line_number + 1]).groupdict()
                    session['iqn'] = line.split()[2]
                availble_targets.append(session)
        return availble_targets

    def _parse_discovery_address(self, iqn):
        '''get an iqn of discovered target and return the discovery ip address
        '''
        # TODO: support multiple discovery addresses
        import re
        discovery_addresses = []
        IQN(iqn)  # make sure it's valid iqn
        cmd = ['iscsiadm', 'list', 'discovery-address']
        process = self._execute_assert_n_log(cmd)
        regex = re.compile('Discovery Address: 'r'(?P<ip>\d+\.\d+\.\d+\.\d+)\:(?P<port>\d+)')
        for line in process.get_stdout().splitlines():
            discovery_addresses.append(regex.search(line).groupdict()['ip'])
        for path in self._parse_discovered_targets():
            if path['iqn'] == iqn:
                if path['dst_ip'] in discovery_addresses:
                    return (path['dst_ip'], path['dst_port'])

    def _parse_availble_sessions(self):
        import re
        availble_sessions = []
        cmd = ['iscsiadm', 'list', 'target', '-v']
        process = self._execute_assert_n_log(cmd)
        output = process.get_stdout().splitlines()
        logger.debug([line.strip() for line in output])
        source_ip_regex = re.compile('IP address \(Local\): 'r'(?P<src_ip>\d+\.\d+\.\d+\.\d+)\:(?P<src_port>\d+)')
        target_ip_regex = re.compile('IP address \(Peer\): 'r'(?P<dst_ip>\d+\.\d+\.\d+\.\d+)\:(?P<dst_port>\d+)')
        for line_number, line in enumerate(output):
            if 'Target: ' in line:
                iqn = line.split()[1]
                IQN(iqn)  # make sure iqn is valid
                for ident_line in range(line_number, len(output)):
                    if 'TPGT: ' in output[ident_line]:
                        uid = output[ident_line].split()[1]
                    if source_ip_regex.search(output[ident_line]):
                        session = source_ip_regex.search(output[ident_line]).groupdict()
                    if target_ip_regex.search(output[ident_line]):
                        session.update(target_ip_regex.search(output[ident_line]).groupdict())
                        session['iqn'] = iqn
                        session['uid'] = uid
                        availble_sessions.append(session)
                        break
                    if re.search('Login Parameters', output[ident_line]):
                        # max search - no point searching after here
                        break
        return availble_sessions

    def get_discovered_targets(self):
        iqn_list = []
        targets = []
        for connectivity in self._parse_discovered_targets():
            iqn_list.append(connectivity['iqn'])
        uniq_iqn = list(set(iqn_list))
        for iqn in uniq_iqn:
            endpoints = []
            if self._parse_discovery_address(iqn) is None:
                continue
            ip_address, port = self._parse_discovery_address(iqn)
            discovery_endpoint = base.Endpoint(ip_address, port)
            for connectivity in self._parse_discovered_targets():
                if connectivity['iqn'] == iqn:
                    endpoints.append(base.Endpoint(connectivity['dst_ip'], connectivity['dst_port']))
            targets.append(base.Target(endpoints, discovery_endpoint, iqn))
        return targets

    def get_source_iqn(self):
        '''return infi.dtypes.iqn type iqn if iscsi initiator file exists
        '''
        import re
        process = self._execute_assert_n_log(['iscsiadm', 'list', 'initiator-node'])
        iqn_line = process.get_stdout().splitlines()[0]
        if re.search(r'Initiator node name', iqn_line):
            iqn = iqn_line.split('Initiator node name: ')[1]
            return IQN(iqn)  # Validate iqn is legal
        else:
            raise RuntimeError("Couldn't find IQN from iscsiadm output, got {!r}".format(process.get_stdout()))

    def reset_source_iqn(self):
        pass

    def set_source_iqn(self, iqn):
        '''receives a string, validates it's an iqn then set it to the host
        NOTE: this restart the iscsi service and may fail active sessions !
        in Solaris, this doesn't save a copy of the old IQN
        '''
        IQN(iqn)   # checks iqn is valid
        old_iqn = self.get_source_iqn()  # check file exist and valid
        execute_assert_success(['iscsiadm', 'modify', 'initiator-node', '-N', iqn])
        logger.info("iqn was replaced from {} to {}".format(old_iqn, iqn))

    def _enable_iscsi_auto_login(self):
        cmd = ['iscsiadm', 'modify', 'discovery', '--sendtargets', 'enable']
        return self._execute_assert_n_log(cmd)

    def _disable_iscsi_auto_login(self):
        cmd = ['iscsiadm', 'modify', 'discovery', '--sendtargets', 'disable']
        return self._execute_assert_n_log(cmd)

    def _modify_target(self, key, value, iqn):
        cmd = ['iscsiadm', 'modify', 'target-param', key, value, str(iqn)]
        self._execute_assert_n_log(cmd)

    def _modify_initiator(self, key, value):
        cmd = ['iscsiadm', 'modify', 'initiator-node', key, value]
        self._execute_assert_n_log(cmd)

    def _clear_auth(self):
        cmd = ['iscsiadm', 'modify', 'initiator-node', '--authentication', 'none']
        self._execute_assert_n_log(cmd)

        for line in self._execute_assert_n_log(['iscsiadm', 'list', 'target-param']).get_stdout().splitlines():
            if not line.startswith('Target'):
                return
            iqn = line.strip().split()[1]
            self._modify_target('--bi-directional-authentication', 'disable', iqn)
            self._modify_target('--authentication', 'none', iqn)

    def _chap_set_password(self, cmd, password):
        # Solaris support chap pass of 12-16 characters
        # potiontal bug when password reset doesn't work
        import pexpect
        from infi.iscsiapi.iscsi_exceptions import ChapPasswordTooLong
        if len(password) > 16:
            raise ChapPasswordTooLong()
        logger.debug("running: {}".format(cmd))
        process = pexpect.spawn(cmd)
        process.expect("Enter secret:")
        process.sendline(password)
        process.expect("Re-enter secret:")
        process.sendline(password)
        logger.debug("password reset finished with exit code {}".format(process.exitstatus))

    def _set_auth(self, auth, iqn):
        def _set_unidirectional_chap():
            self._modify_target('--bi-directional-authentication', 'disable', iqn)
            self._modify_target('--authentication', 'chap', iqn)
            self._modify_initiator('--CHAP-name', auth.get_inbound_username())
            self._chap_set_password('iscsiadm modify initiator-node --CHAP-secret', auth.get_inbound_secret())

        def _set_bidirectional_chap():
            self._modify_target('--bi-directional-authentication', 'enable', iqn)
            self._modify_target('--CHAP-name', auth.get_outbound_username(), iqn)
            cmd = 'iscsiadm modify target-param --CHAP-secret {}'.format(iqn)
            self._chap_set_password(cmd, auth.get_outbound_secret())

        if isinstance(auth, iscsiapi_auth.ChapAuth):
            _set_unidirectional_chap()
        elif isinstance(auth, iscsiapi_auth.MutualChapAuth):
            _set_unidirectional_chap()
            _set_bidirectional_chap()
        elif isinstance(auth, iscsiapi_auth.NoAuth):
            self._modify_target('--bi-directional-authentication', 'disable', iqn)
            self._modify_target('--authentication', 'none', iqn)

    def discover(self, ip_address, port=3260):
        '''initiate discovery and returns a list of dicts which contain all available targets
        '''
        from .iscsi_exceptions import DiscoveryFailed
        endpoints = []
        self._modify_initiator('--authentication', 'none')
        args = ['iscsiadm', 'add', 'discovery-address', str(ip_address) + ':' + str(port)]
        self._execute_assert_n_log(args)
        for target_connectivity in self._parse_discovered_targets():
            if target_connectivity['dst_ip'] == ip_address:
                iqn = target_connectivity['iqn']
                break
        else:
            raise DiscoveryFailed()
        for target_connectivity in self._parse_discovered_targets():
            if iqn == target_connectivity['iqn']:
                endpoints.append(base.Endpoint(target_connectivity['dst_ip'], target_connectivity['dst_port']))
        return base.Target(endpoints, base.Endpoint(ip_address, port), iqn)

    def undiscover(self, target=None):
        '''logout from everything and delete all discovered target if target=None otherwise delete only the target
        discovery endpoints
        '''
        import re
        if target:
            ip_address = target.get_discovery_endpoint().get_ip_address()
            self._execute_n_log(['iscsiadm', 'remove', 'discovery-address', ip_address])
        else:
            cmd = ['iscsiadm', 'list', 'discovery-address']
            process = self._execute_assert_n_log(cmd)
            regex = re.compile('Discovery Address: 'r'(?P<ip>\d+\.\d+\.\d+\.\d+)\:(?P<port>\d+)')
            for line in process.get_stdout().splitlines():
                self._execute_n_log(['iscsiadm', 'remove', 'discovery-address', regex.search(line).groupdict()['ip']])

    def login(self, target, endpoint, auth=None, num_of_connections=1):
        raise NotImplemented("In Solaris login is supported only to all available endpoints\n" +
                             "Therefore, login to a single endpoint couldn't be implemented")

    def login_all(self, target, auth=None):
        if auth is None:
            auth = iscsiapi_auth.NoAuth()
        logger.info("login_all in Solaris login to all available Targets !")
        logger.warn("Performing login in Solaris disconnect momentarily all previous sessions from all targets")
        self._set_auth(auth, target.get_iqn())
        self._disable_iscsi_auto_login()
        self._enable_iscsi_auto_login()
        return self.get_sessions(target=target)

    def logout(self, session):
        raise NotImplemented("Logout from a single session isn't supported in Solaris")

    def logout_all(self, target):
        logger.warn("Performing logout in Solaris disconnect momentarily all sessions from all targets")
        self.undiscover(target)
        self._disable_iscsi_auto_login()
        self._enable_iscsi_auto_login()

    def get_sessions(self, target=None):
        '''receive a target or None and return a list of all available sessions
        '''
        # TODO: add HCT to session
        def get_sessions_for_target(target):
            from infi.dtypes.hctl import HCT
            target_sessions = []
            for session in self._parse_availble_sessions():
                if session['iqn'] == target.get_iqn():
                    hct = HCT(session['src_ip'], 0, session['dst_ip'])
                    target_sessions.append(base.Session(target, base.Endpoint(session['dst_ip'], session['dst_port']),
                                 session['src_ip'], self.get_source_iqn(), session['uid'], hct))
            return target_sessions

        if target:
            return get_sessions_for_target(target)
        else:
            sessions = []
            targets = self.get_discovered_targets()
            for target in targets:
                sessions.extend(get_sessions_for_target(target))
            return sessions

    def rescan(self):
        '''does nothing in Solaris
        '''
        logger.info("Someone just initiated iscsi rescan, In Solaris it does nothing")
        pass

class SolarisSoftwareInitiator(base.SoftwareInitiator):
    def is_installed(self):
        ''' Return True if iSCSI initiator sw is installed otherwise return False
        '''
        if 'solaris' in get_platform_string():
            process1 = execute('pkginfo', '-q', 'SUNWiscsir')
            process2 = execute('pkginfo', '-q', 'SUNWiscsiu')
            if process1.get_returncode() == process2.get_returncode() == 0:
                return True
            else:
                return False

    def install(self):
        # Not installing iscsi utils on solaris suppose to come by default
        pass

    def uninstall(self):
        # Does nothing, not implemented in pkgmgr
        pass
